<?php

namespace Drupal\commerce_paybox_payment\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_paybox_payment\Controller\PaymentController;
use Drupal\commerce_paybox_payment\Services\PayboxDirectApiServiceInterface;
use Drupal\commerce_paybox_payment\Services\PbxCmdRefHelperInterface;
use Drupal\commerce_paybox_payment\SignatureChecker;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_price\Price;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageManager;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Config\ConfigFactory;

/**
 * Paybox Commerce Payment Gateway.
 *
 * @CommercePaymentGateway(
 *   id = "paybox",
 *   label = @Translation("Paybox"),
 *   display_label = @Translation("Paybox"),
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_paybox_payment\PluginForm\RedirectPayboxForm",
 *     "add-payment" = "Drupal\commerce_paybox_payment\PluginForm\PayboxPaymentAddForm",
 *   },
 *   payment_method_types = {"credit_card"},
 * )
 */
final class Paybox extends OffsitePaymentGatewayBase implements SupportsRefundsInterface {
  use LoggerChannelTrait;

  private const SERVICES = [
    'paybox' => 'Paybox (by Verifone)',
    'up2pay' => 'Up2pay (by Crédit Agricole)',
  ];

  private const PROTOCOL_VERSIONS = [
    '00103' => 'Paybox Direct / Up2pay e-transactions Access',
    '00104' => 'Paybox Direct Plus / Up2pay e-transactions Premium',
  ];

  private const CURRENCY_CODE_MAPPING = [
    '978' => 'EUR',
    '840' => 'USD',
    '826' => 'GBP',
    '032' => 'ARS',
    '040' => 'ATS',
    '036' => 'AUD',
    '056' => 'BEF',
    '986' => 'BRL',
    '124' => 'CAD',
    '756' => 'CHF',
    '280' => 'DEM',
    '208' => 'DKK',
    '246' => 'FIM',
    '300' => 'GRD',
    '392' => 'JPY',
    '116' => 'KHR',
    '410' => 'KRW',
    '484' => 'MXN',
    '578' => 'NOK',
    '554' => 'NZD',
    '752' => 'SEK',
    '702' => 'SGD',
    '792' => 'TRL',
    '901' => 'TWD',
  ];

  private const LANGUAGE_CODE_MAPPING = [
    'fr' => 'fra',
    'it' => 'ita',
    'sv' => 'swe',
    'en' => 'gbr',
    'de' => 'deu',
    'pt-pt' => 'prt',
    'es' => 'esp',
    'nl' => 'nld',
  ];

  private const ENTITY_ID_SEPARATOR = [
    '-' => '-',
    '_' => '_',
  ];

  /**
   * Psr\Log\LoggerInterface definition.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * LanguageManager Service.
   *
   * @var \Drupal\Core\Language\LanguageManager
   */
  protected $languageManager;

  /**
   * Drupal\Core\Config\Config Service.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $globalConfig;

  /**
   * Drupal\commerce_paybox_payment\Services\PbxCmdRefHelperInterface Service.
   *
   * @var \Drupal\commerce_paybox_payment\Services\PbxCmdRefHelperInterface
   */
  protected $pbxCmdRefHelper;

  /**
   * Drupal\commerce_paybox_payment\Services\PayboxDirectApiService Service.
   *
   * @var \Drupal\commerce_paybox_payment\Services\PayboxDirectApiServiceInterface
   */
  protected $api;

  /**
   * Paybox constructor.
   *
   * @param array $configuration
   *   Plugin configuration.
   * @param string $plugin_id
   *   Plugin id.
   * @param mixed $plugin_definition
   *   Plugin definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity Type Manager Service.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   Payment Type Manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   Payment Method Type Manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   Time Service.
   * @param \Drupal\Core\Language\LanguageManager $languageManager
   *   LanguageManager Service.
   * @param \Drupal\Core\Config\ConfigFactory $configFactory
   *   The Config Factory Service.
   * @param \Drupal\commerce_paybox_payment\Services\PbxCmdRefHelperInterface $pbxCmdRefHelper
   *   The Paybox Command Ref Helper Service.
   * @param \Drupal\commerce_paybox_payment\Services\PayboxDirectApiServiceInterface $payboxDirectApiService
   *   The Paybox Direct API Service.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    PaymentTypeManager $payment_type_manager,
    PaymentMethodTypeManager $payment_method_type_manager,
    TimeInterface $time,
    LanguageManager $languageManager,
    ConfigFactory $configFactory,
    PbxCmdRefHelperInterface $pbxCmdRefHelper,
    PayboxDirectApiServiceInterface $payboxDirectApiService
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);
    $this->logger = $this->getLogger('commerce_paybox_payment');
    $this->languageManager = $languageManager;
    $this->globalConfig = $configFactory->getEditable('commerce_paybox_payment.settings');
    $this->pbxCmdRefHelper = $pbxCmdRefHelper;
    $this->api = $payboxDirectApiService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('language_manager'),
      $container->get('config.factory'),
      $container->get('commerce_paybox_payment.pbx_cmd_ref_helper'),
      $container->get('commerce_paybox_payment.paybox_direct_api'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'service' => '',
      'protocol_version' => '',
      'merchant_id' => '',
      'rank' => '',
      'login' => '',
      'secret_key' => '',
      'server_url' => '',
      'currency_code' => '',
      'payment_diff' => '',
      'entity_id_mapping' => 'payment_id',
      'entity_id_separator' => '_',
      'use_secondary_ppps_endpoint' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['service'] = [
      '#type' => 'select',
      '#title' => $this->t('Service'),
      '#options' => self::SERVICES,
      '#default_value' => $this->configuration['service'],
      '#required' => TRUE,
    ];
    $form['merchant_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Site'),
      '#size' => 20,
      '#default_value' => $this->configuration['merchant_id'],
      '#required' => TRUE,
    ];
    $form['rank'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rank'),
      '#size' => 20,
      '#default_value' => $this->configuration['rank'],
      '#required' => TRUE,
    ];
    $form['login'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Login'),
      '#size' => 20,
      '#default_value' => $this->configuration['login'],
      '#required' => TRUE,
    ];
    $form['secret_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('HMAC Secret key'),
      '#default_value' => $this->configuration['secret_key'],
      '#required' => TRUE,
    ];
    $form['server_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('The Paybox payment server URL'),
      '#description' => $this->t('The absolute URL to the Paybox payment server to process transactions.'),
      '#size' => 100,
      '#default_value' => $this->configuration['server_url'],
      '#required' => TRUE,
    ];
    $form['protocol_version'] = [
      '#type' => 'select',
      '#title' => $this->t('PPPS protocol version'),
      '#options' => self::PROTOCOL_VERSIONS,
      '#default_value' => $this->configuration['protocol_version'],
      '#required' => TRUE,
    ];
    $form['use_secondary_ppps_endpoint'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use secondary PPPS endpoint'),
      '#description' => $this->t('Production server provides two endpoint for API endpoint, primary and secondary.') .
      '<br />' . $this->t('Any doubts, use primary (by not checking this box).'),
      '#default_value' => $this->configuration['use_secondary_ppps_endpoint'],
    ];
    $form['currency_code'] = [
      '#type' => 'select',
      '#title' => $this->t('Currency'),
      '#options' => self::CURRENCY_CODE_MAPPING,
      '#default_value' => $this->configuration['currency_code'],
    ];
    $form['payment_diff'] = [
      '#type' => 'textfield',
      '#attributes' => [
        ' type' => 'number',
      ],
      '#title' => $this->t('Deferred payment'),
      '#default_value' => $this->configuration['payment_diff'],
    ];
    $form['entity_id_mapping'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity ID mapping'),
      '#description' => $this->t('This will be displayed in the Paybox/Up2Pay backoffice.'),
      '#options' => [
        'payment_id' => $this->t('Payment ID'),
        'payment_and_order_ids' => $this->t('Payment and order IDs'),
      ],
      '#default_value' => $this->globalConfig->get('entity_id_mapping'),
      '#required' => TRUE,
    ];
    $form['entity_id_separator'] = [
      '#type' => 'select',
      '#title' => $this->t('Separator'),
      '#description' => $this->t('Make sure your order ID is not containing the separator, it could break the mapping.') .
      '<br />' . $this->t('Any doubts, please select <strong>Payment ID</strong> option.'),
      '#options' => self::ENTITY_ID_SEPARATOR,
      '#default_value' => $this->globalConfig->get('entity_id_separator'),
      '#required' => FALSE,
      '#states' => [
        'visible' => [
          ':input[name="configuration[paybox][entity_id_mapping]"]' => [
            'value' => 'payment_and_order_ids',
          ],
        ],
      ],
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $values = $form_state->getValue($form['#parents']);
    $this->configuration['service'] = $values['service'];
    $this->configuration['merchant_id'] = $values['merchant_id'];
    $this->configuration['rank'] = $values['rank'];
    $this->configuration['login'] = $values['login'];
    $this->configuration['secret_key'] = $values['secret_key'];
    $this->configuration['server_url'] = $values['server_url'];
    $this->configuration['protocol_version'] = $values['protocol_version'];
    $this->configuration['use_secondary_ppps_endpoint'] = $values['use_secondary_ppps_endpoint'];
    $this->configuration['currency_code'] = $values['currency_code'];
    $this->configuration['payment_diff'] = $values['payment_diff'];

    $this->globalConfig->set('entity_id_mapping', $values['entity_id_mapping']);
    $this->globalConfig->set('entity_id_separator', $values['entity_id_separator']);
    $this->globalConfig->save();
  }

  /**
   * Construct Paybox request data.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   Order entity.
   * @param \Drupal\commerce_payment\Entity\PaymentInterface $payment
   *   Payment entity.
   *
   * @return array
   *   Request data build.
   */
  public function request(OrderInterface $order, PaymentInterface $payment) {
    $plugin_config = $this->getConfiguration();

    // Save payment to have reference to pass to bank site.
    $payment->save();

    // IPN return callback URL (to override platform config).
    $ipn_return_url = Url::fromRoute('commerce_paybox_payment.payment_api_return', [
      'commerce_order' => $order->id(),
    ])->setAbsolute(TRUE)->toString();

    $paymentAmount = (float) $payment->getAmount()->getNumber() * 100;
    $data = [
      'PBX_SITE' => $plugin_config['merchant_id'],
      'PBX_RANG' => $plugin_config['rank'],
      'PBX_IDENTIFIANT' => $plugin_config['login'],
      'PBX_TOTAL' => str_pad((string) $paymentAmount, 3, '0', STR_PAD_LEFT),
      'PBX_DEVISE' => $plugin_config['currency_code'],
      'PBX_CMD' => $this->pbxCmdRefHelper->createPbxCmdValueFromPayment($payment),
      'PBX_PORTEUR' => $order->getEmail(),
      'PBX_RETOUR' => 'Mt:M;Ref:R;Transaction:S;Appel:T;Date:Q;Auto:A;Error:E;Signature:K',
      'PBX_HASH' => 'SHA512',
      'PBX_TIME' => date('c', $order->getCreatedTime()),
      'PBX_REPONDRE_A' => $ipn_return_url,
    ];

    $language = $this->languageManager->getCurrentLanguage()->getId();
    if (!empty(self::LANGUAGE_CODE_MAPPING[$language])) {
      $data['PBX_LANGUE'] = strtoupper(self::LANGUAGE_CODE_MAPPING[$language]);
    }

    if (!empty($plugin_config['payment_diff'])) {
      $data['PBX_DIFF'] = (int) $plugin_config['payment_diff'];
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    $paymentId = $this->pbxCmdRefHelper->extractPaymentIdFromRef($request);
    $payment = $this->entityTypeManager->getStorage('commerce_payment')->load($paymentId);

    if (!$payment instanceof PaymentInterface) {
      throw new PaymentGatewayException('Payment cannot be found');
    }

    if (1 !== SignatureChecker::checkSignature($request)) {
      throw new PaymentGatewayException('Invalid Signature');
    }

    if ('completed' === $payment->getState()->getId()) {
      throw new PaymentGatewayException('Payment is already completed');
    }

    $payment->setState('voided');
    $payment->save();
    parent::onCancel($order, $request);
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $paymentId = $this->pbxCmdRefHelper->extractPaymentIdFromRef($request);
    $transactionId = $request->query->get('Transaction');
    $appelId = $request->query->get('Appel');
    $transactionDate = $request->query->get('Date');
    $payment = $this->entityTypeManager->getStorage('commerce_payment')->load($paymentId);

    if (!$payment instanceof PaymentInterface) {
      throw new PaymentGatewayException('Payment cannot be found');
    }

    if (1 !== SignatureChecker::checkSignature($request)) {
      throw new PaymentGatewayException('Invalid Signature');
    }

    if ('completed' === $payment->getState()->getId()) {
      throw new PaymentGatewayException('Payment is already completed');
    }

    $returnAmount = $request->query->get('Mt');
    $paymentAmount = (float) $payment->getAmount()->getNumber() * 100;
    $paymentAmount = str_pad((string) $paymentAmount, 3, '0', STR_PAD_LEFT);
    $error = $request->query->get('Error');
    if ($returnAmount !== $paymentAmount) {
      throw new PaymentGatewayException('Charged amount not equal to order amount');
    }
    if (!empty($error) && PaymentController::EMPTY_ERROR !== $error  && PaymentController::PENDING_CODE !== $error) {
      throw new PaymentGatewayException('An error has occurred during payment');
    }

    if (PaymentController::PENDING_CODE === $error) {
      $payment->setState('pending');
      $payment->set('remote_id', $transactionId);
      $payment->save();
      parent::onReturn($order, $request);
      return;
    }

    $payment->setState('completed');
    $payment->set('remote_id', Json::encode([
      'TRANS' => $transactionId,
      'APPEL' => $appelId,
    ]));
    $payment->set('authorized', strtotime('today ' . $transactionDate));
    $payment->save();
    parent::onReturn($order, $request);
  }

  /**
   * {@inheritdoc}
   */
  public function canRefundPayment(PaymentInterface $payment) {
    // Only completed or partially refunded payments can be refunded.
    return in_array($payment->getState()->getId(), [
      'completed',
      'partially_refunded',
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);

    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $this->assertRefundAmount($payment, $amount);

    // Perform the refund request here, throw an exception if it fails.
    try {
      $this->api->refund($payment, $amount);
    }
    catch (\Exception $e) {
      $this->logger->log('error', 'An error happened while calling Paybox API: ' . $e->getMessage());
      throw new PaymentGatewayException('An error happened while calling Paybox API');
    }

    // Determine whether payment has been fully or partially refunded.
    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

}
